import multiprocessing as mp
import numpy as np
# import warnings
# import sys


def ensemble_sampler_cm_graph(outfile_name, x, cpu_n=2, seed=None):
    """Produce a single undirected binary graph after UBCM.
        The graph is saved as an edges list on a file.

    :param outfile_name: Name of the file where to save the graph on.
    :type outfile_name: str
    :param x: UBCM solution
    :type x: numpy.ndarray
    :param cpu_n: Number of cpus to use, defaults to 2
    :type cpu_n: int, optional
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: The `outfile_name` input parameter
    :rtype: str
    """
    if seed is not None:
        np.random.seed(seed)

    inds = np.arange(len(x))

    # put together inputs for pool
    iter_ = iter(
        ((i, xi), (j, xj), np.random.randint(0, 1000000))
        for i, xi in zip(inds, x)
        for j, xj in zip(inds, x)
        if i < j)

    # compute existing edges
    with mp.Pool(processes=cpu_n) as pool:
        edges_list = pool.starmap(is_a_link_cm, iter_)

    # removing None
    edges_list[:] = (value for value in edges_list if value is not None)

    # edgelist writing
    with open(outfile_name, "w") as outfile:
        outfile.write(
            "".join(
                "{} {}\n".format(str(i), str(j))
                for (i, j) in edges_list)
            )

    return outfile_name


def ensemble_sampler_ecm_graph(outfile_name, x, y, cpu_n=2, seed=None):
    """Produce a single undirected weighted graph after ECM.
        The graph is saved as an edges list on a file.
        The function is parallelyzed.

    :param outfile_name: Name of the file where to save the graph on.
    :type outfile_name: str
    :param x: ECM degrees solution
    :type x: numpy.ndarray
    :param y: ECM strengths solution
    :type y: numpy.ndarray
    :param cpu_n: Number of cpus to use, defaults to 2
    :type cpu_n: int, optional
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: The `outfile_name` input parameter
    :rtype: str
    """
    if seed is not None:
        np.random.seed(seed)

    inds = np.arange(len(x))

    # put together inputs for pool
    iter_ = iter(
        ((i, xi, yi), (j, xj, yj), np.random.randint(0, 1000000))
        for i, xi, yi in zip(inds, x, y)
        for j, xj, yj in zip(inds, x, y)
        if i < j)

    # compute existing edges
    with mp.Pool(processes=cpu_n) as pool:
        edges_list = pool.starmap(is_a_link_ecm, iter_)

    # removing None
    edges_list[:] = (value for value in edges_list if value is not None)

    # edgelist writing
    with open(outfile_name, "w") as outfile:
        outfile.write(
            "".join(
                "{} {} {}\n".format(str(i), str(j), str(w))
                for (i, j, w) in edges_list)
            )

    return outfile_name


def ensemble_sampler_dcm_graph(outfile_name, x, y, cpu_n=2, seed=None):
    """Produce a single directed binary graph after DBCM.
        The graph is saved as an edges list on a file.
        The function is parallelyzed.

    :param outfile_name: Name of the file where to save the graph on.
    :type outfile_name: str
    :param x: DBCM out-degrees parameters solution
    :type x: numpy.ndarray
    :param y: DBCM in-degrees parameters solution
    :type y: numpy.ndarray
    :param cpu_n: Number of cpus to use, defaults to 2
    :type cpu_n: int, optional
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: The `outfile_name` input parameter
    :rtype: str
    """
    if seed is not None:
        np.random.seed(seed)

    inds = np.arange(len(x))

    # put together inputs for pool
    iter_ = iter(((i, xi), (j, yj), np.random.randint(0, 1000000))
                 for i, xi in zip(inds, x)
                 for j, yj in zip(inds, y)
                 if i != j)

    # compute existing edges
    with mp.Pool(processes=cpu_n) as pool:
        edges_list = pool.starmap(is_a_link_dcm, iter_)

    # removing None
    edges_list[:] = (value for value in edges_list if value is not None)

    # edgelist writing
    with open(outfile_name, "w") as outfile:
        outfile.write(
            "".join(
                "{} {}\n".format(str(i), str(j))
                for (i, j) in edges_list)
            )

    return outfile_name


def ensemble_sampler_decm_graph(
        outfile_name,
        a_out, a_in,
        b_out, b_in,
        cpu_n=2,
        seed=None):
    """Produce a single directed weighted graph after DECM.
        The graph is saved as an edges list on a file.
        The function is parallelyzed.

    :param outfile_name: Name of the file where to save the graph on.
    :type outfile_name: str
    :param a_out: DECM out-degrees parameters solution
    :type a_out: numpy.ndarray
    :param a_in: DECM in-degrees parameters solution
    :type a_in: numpy.ndarray
    :param b_out: DECM out-strengths parameters solution
    :type b_out: numpy.ndarray
    :param b_in: DECM in-strengths parameters solution
    :type b_in: numpy.ndarray
    :param cpu_n: Number of cpus to use, defaults to 2
    :type cpu_n: int, optional
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: The `outfile_name` input parameter
    :rtype: str
    """
    if seed is not None:
        np.random.seed(seed)

    inds = np.arange(len(a_out))

    # put together inputs for pool
    iter_ = iter((
                    (i, a_out_i, b_out_i),
                    (j, a_in_j, b_in_j),
                    np.random.randint(0, 1000000))
                 for i, a_out_i, b_out_i in zip(inds, a_out, b_out)
                 for j, a_in_j, b_in_j in zip(inds, a_in, b_in)
                 if i != j)

    # compute existing edges
    with mp.Pool(processes=cpu_n) as pool:
        edges_list = pool.starmap(is_a_link_decm, iter_)

    # removing None
    edges_list[:] = (value for value in edges_list if value is not None)

    # edgelist writing
    with open(outfile_name, "w") as outfile:
        outfile.write(
            "".join(
                "{} {} {}\n".format(str(i), str(j), str(w))
                for (i, j, w) in edges_list)
            )

    return outfile_name


def ensemble_sampler_crema_ecm_det_graph(
        outfile_name,
        beta,
        adj,
        cpu_n=2,
        seed=None):
    """Produce a single undirected weighted graph after CReMa,
        given a fixed, complete network topology.
        The graph is saved as an edges list on a file.
        The function is parallelyzed.

    :param outfile_name: Name of the file where to save the graph on.
    :type outfile_name: str
    :param beta: CReMa solution
    :type beta: numpy.ndarray
    :param adj: Edges list: out-nodes array, in-nodes array and links weights.
    :type adj: (numpy.ndarray, numpy.ndarray, numpy.ndarray)
    :param cpu_n: Number of cpus to use, defaults to 2
    :type cpu_n: int, optional
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: The `outfile_name` input parameter
    :rtype: str
    """
    if seed is not None:
        np.random.seed(seed)

    (row_inds, col_inds, weigths_value) = adj
    del weigths_value

    # put together inputs for pool
    iter_ = iter(
        ((i, beta[i]), (j, beta[j]), np.random.randint(0, 1000000))
        for i, j in zip(row_inds, col_inds)
    )

    # compute existing edges
    with mp.Pool(processes=cpu_n) as pool:
        edges_list = pool.starmap(is_a_link_crema_ecm_det, iter_)

    # edgelist writing
    with open(outfile_name, "w") as outfile:
        outfile.write(
            "".join(
                "{} {} {}\n".format(str(i), str(j), str(w))
                for (i, j, w) in edges_list)
            )

    return outfile_name


def ensemble_sampler_crema_ecm_prob_graph(
        outfile_name,
        beta,
        adj,
        cpu_n=2,
        seed=None):
    """Produce a single undirected weighted graph after CReMa,
        given a probabilistic network topology.
        The graph is saved as an edges list on a file.
        The function is parallelyzed.

    :param outfile_name: Name of the file where to save the graph on.
    :type outfile_name: str
    :param beta: CReMa solution
    :type beta: numpy.ndarray
    :param adj: Edges list:
         out-nodes array, in-nodes array and links probabilities.
    :type adj: (numpy.ndarray, numpy.ndarray, numpy.ndarray)
    :param cpu_n: Number of cpus to use, defaults to 2
    :type cpu_n: int, optional
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: The `outfile_name` input parameter
    :rtype: str
    """
    if seed is not None:
        np.random.seed(seed)

    (row_inds, col_inds, weigths_value) = adj

    # put together inputs for pool
    iter_ = iter(
        ((i, beta[i]), (j, beta[j]), w_prob, np.random.randint(0, 1000000))
        for i, j, w_prob in zip(row_inds, col_inds, weigths_value)
    )

    # compute existing edges
    with mp.Pool(processes=cpu_n) as pool:
        edges_list = pool.starmap(is_a_link_crema_ecm_prob, iter_)

    # removing None
    edges_list[:] = (value for value in edges_list if value is not None)

    # edgelist writing
    with open(outfile_name, "w") as outfile:
        outfile.write(
            "".join(
                "{} {} {}\n".format(str(i), str(j), str(w))
                for (i, j, w) in edges_list)
            )

    return outfile_name


def ensemble_sampler_crema_sparse_ecm_prob_graph(
        outfile_name,
        beta,
        adj,
        cpu_n=2,
        seed=None):
    """Produce a single undirected weighted graph after CReMa,
        given a probabilistic network topology.
        The graph is saved as an edges list on a file.
        The function is parallelyzed.

    :param outfile_name: Name of the file where to save the graph on.
    :type outfile_name: str
    :param beta: CReMa solution
    :type beta: numpy.ndarray
    :param adj: Edges list:
         out-nodes array, in-nodes array and links probabilities.
    :type adj: (numpy.ndarray, numpy.ndarray, numpy.ndarray)
    :param cpu_n: Number of cpus to use, defaults to 2
    :type cpu_n: int, optional
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: The `outfile_name` input parameter
    :rtype: str
    """
    if seed is not None:
        np.random.seed(seed)

    x = adj[0]
    n = len(x)

    # put together inputs for pool
    iter_ = iter(
        ((i, beta[i], x[i]), (j, beta[j], x[j]), np.random.randint(0, 1000000))
        for i in range(n)
        for j in range(n)
        if i != j
    )

    # compute existing edges
    with mp.Pool(processes=cpu_n) as pool:
        edges_list = pool.starmap(is_a_link_crema_sparse_ecm_prob, iter_)

    # removing None
    edges_list[:] = (value for value in edges_list if value is not None)

    # debug
    # print(edges_list)

    # edgelist writing
    with open(outfile_name, "w") as outfile:
        outfile.write(
            "".join(
                "{} {} {}\n".format(str(i), str(j), str(w))
                for (i, j, w) in edges_list)
            )

    return outfile_name


def ensemble_sampler_crema_decm_prob_graph(
        outfile_name,
        beta,
        adj,
        cpu_n=2,
        seed=None):
    """Produce a single directed weighted graph after CReMa,
        given a probabilistic network topology.
        The graph is saved as an edges list on a file.
        The function is parallelyzed.

    :param outfile_name: Name of the file where to save the graph on.
    :type outfile_name: str
    :param beta: CReMa solution
    :type beta: numpy.ndarray
    :param adj: Edges list:
         out-nodes array, in-nodes array and links probabilities.
    :type adj: (numpy.ndarray, numpy.ndarray, numpy.ndarray)
    :param cpu_n: Number of cpus to use, defaults to 2
    :type cpu_n: int, optional
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: The `outfile_name` input parameter
    :rtype: str
    """
    if seed is not None:
        np.random.seed(seed)

    b_out, b_in = beta

    (row_inds, col_inds, weigths_value) = adj

    # put together inputs for pool
    iter_ = iter(
        ((i, b_out[i]), (j, b_in[j]), w_prob, np.random.randint(0, 1000000))
        for i, j, w_prob in zip(row_inds, col_inds, weigths_value)
    )

    # compute existing edges
    with mp.Pool(processes=cpu_n) as pool:
        edges_list = pool.starmap(is_a_link_crema_decm_prob, iter_)

    # debug
    # print(edges_list)

    # removing None
    edges_list[:] = (value for value in edges_list if value is not None)

    # edgelist writing
    with open(outfile_name, "w") as outfile:
        outfile.write(
            "".join(
                "{} {} {}\n".format(str(i), str(j), str(w))
                for (i, j, w) in edges_list)
            )

    return outfile_name


def ensemble_sampler_crema_decm_det_graph(
        outfile_name,
        beta,
        adj,
        cpu_n=2,
        seed=None):
    """Produce a single directed weighted graph after CReMa,
        given a fixed, complete network topology.
        The graph is saved as an edges list on a file.
        The function is parallelyzed.

    :param outfile_name: Name of the file where to save the graph on.
    :type outfile_name: str
    :param beta: CReMa solution
    :type beta: numpy.ndarray
    :param adj: Edges list:
         out-nodes array, in-nodes array and links weights.
    :type adj: (numpy.ndarray, numpy.ndarray, numpy.ndarray)
    :param cpu_n: Number of cpus to use, defaults to 2
    :type cpu_n: int, optional
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: The `outfile_name` input parameter
    :rtype: str
    """
    if seed is not None:
        np.random.seed(seed)

    b_out, b_in = beta

    (row_inds, col_inds, weigths_value) = adj
    del weigths_value

    # put together inputs for pool
    iter_ = iter(
        ((i, b_out[i]), (j, b_in[j]), np.random.randint(0, 1000000))
        for i, j in zip(row_inds, col_inds)
    )

    # compute existing edges
    with mp.Pool(processes=cpu_n) as pool:
        edges_list = pool.starmap(is_a_link_crema_decm_det, iter_)

    # debug
    # print(edges_list)

    # edgelist writing
    with open(outfile_name, "w") as outfile:
        outfile.write(
            "".join(
                "{} {} {}\n".format(str(i), str(j), str(w))
                for (i, j, w) in edges_list)
            )

    return outfile_name


def ensemble_sampler_crema_sparse_decm_prob_graph(
        outfile_name,
        beta,
        adj,
        cpu_n=2,
        seed=None):
    """Produce a single directed weighted graph after the
        sparse version of CReMa, given a probabilistic network topology.
        The graph is saved as an edges list on a file.
        The function is parallelyzed.

    :param outfile_name: Name of the file where to save the graph on.
    :type outfile_name: str
    :param beta: CReMa solution
    :type beta: numpy.ndarray
    :param adj: Tuple of DBCM out-degrees parameters and
        in-degrees parameters solution.
    :type adj: (numpy.ndarray, numpy.ndarray)
    :param cpu_n: Number of cpus to use, defaults to 2
    :type cpu_n: int, optional
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: The `outfile_name` input parameter
    :rtype: str
    """
    if seed is not None:
        np.random.seed(seed)

    b_out, b_in = beta
    x, y = adj
    n = len(x)

    # put together inputs for pool
    iter_ = iter(
        (
            (i, b_out[i], x[i]),
            (j, b_in[j], y[j]),
            np.random.randint(0, 1000000))
        for i in range(n)
        for j in range(n)
        if i != j
    )

    # compute existing edges
    with mp.Pool(processes=cpu_n) as pool:
        edges_list = pool.starmap(is_a_link_crema_sparse_decm_prob, iter_)

    # removing None
    edges_list[:] = (value for value in edges_list if value is not None)

    # edgelist writing
    with open(outfile_name, "w") as outfile:
        outfile.write(
            "".join(
                "{} {} {}\n".format(str(i), str(j), str(w))
                for (i, j, w) in edges_list)
            )

    return outfile_name


def is_a_link_cm(args_1, args_2, seed=None):
    """The function randomly returns an undirected link
        between two given nodes after the UBCM.

    :param args_1: Tuple containing node 1 and related parameter solution.
    :type args_1: (int, float)
    :param args_2: Tuple containing node 2 and related parameter solution.
    :type args_2: (int, float)
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: If the links exists returns the couple of nodes,
        otherwise returns None
    :rtype: (int, int)
    """
    if seed is not None:
        np.random.seed(seed)
    (i, xi) = args_1
    (j, xj) = args_2
    p = np.random.random_sample()
    xij = xi*xj
    p_ensemble = xij/(1 + xij)
    if p < p_ensemble:
        return (i, j)


def is_a_link_ecm(args_1, args_2, seed=None):
    """The function randomly returns an undirected link with related weight
        between two given nodes after the ECM.

    :param args_1: Tuple containing node 1, degree parameter solution
        and strength parameter solution.
    :type args_1: (int, float, float)
    :param args_2: Tuple containing node 2, degree parameter solution
        and strength parameter solution.
    :type args_2: (int, float, float)
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: If the links exists returns the triplet of nodes and weight,
        otherwise returns None
    :rtype: (int, int, float)
    """
    # q-ensemble source: "Unbiased sampling of network ensembles", WUNs
    if seed is not None:
        np.random.seed(seed)
    (i, xi, yi) = args_1
    (j, xj, yj) = args_2
    p = np.random.random_sample()
    xij = xi*xj
    yij = yi*yj
    p_ensemble = xij*yij/(1 - yij + xij*yij)
    if p < p_ensemble:
        q_ensemble = yij
        w = np.random.geometric(1-q_ensemble)
        return (i, j, w)


def is_a_link_dcm(args_1, args_2, seed=None):
    """The function randomly returns a directed link
        between two given nodes after the DBCM.

    :param args_1: Tuple containing out node and out-degree parameter solution.
    :type args_1: (int, float)
    :param args_2: Tuple containing in node and in-degree parameter solution.
    :type args_2: (int, float)
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: If the links exists returns the couple of nodes,
        otherwise returns None
    :rtype: (int, int)
    """
    if seed is not None:
        np.random.seed(seed)
    (i, xi) = args_1
    (j, yj) = args_2
    p = np.random.random_sample()
    tmp = xi*yj
    p_ensemble = tmp/(1 + tmp)
    if p < p_ensemble:
        return (i, j)


def is_a_link_decm(args_1, args_2, seed=None):
    """The function randomly returns an directed link with related weight
        between two given nodes after the DECM.

    :param args_1: Tuple containing out node, out-degree parameter solution
        and out-strength parameter solution.
    :type args_1: (int, float, float)
    :param args_2: Tuple containing in node, in-degree parameter solution
        and in-strength parameter solution.
    :type args_2: (int, float, float)
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: If the links exists returns the triplet of nodes and weight,
        otherwise returns None
    :rtype: (int, int, float)
    """
    # q-ensemble source:
    # "Fast and scalable resolution of the likelihood maximization problem
    # for Exponential Random Graph models"
    if seed is not None:
        np.random.seed(seed)
    (i, a_out_i, b_out_i) = args_1
    (j, a_in_j, b_in_j) = args_2
    p = np.random.random_sample()
    aij = a_out_i * a_in_j
    bij = b_out_i * b_in_j
    p_ensemble = aij*bij/(1 - bij + aij*bij)
    if p < p_ensemble:
        q_ensemble = bij
        w = np.random.geometric(1-q_ensemble)
        return (i, j, w)


def is_a_link_crema_ecm_det(args_1, args_2, seed=None):
    """The function randomly returns an undirected link with related weight
        between two given nodes after the CReMa undirected
        for a deterministic topology.

    :param args_1: Tuple containing node 1, and strength parameter solution.
    :type args_1: (int, float)
    :param args_2: Tuple containing node 2, and strength parameter solution.
    :type args_2: (int, float)
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: If the links exists returns the triplet of nodes and weight,
        otherwise returns None
    :rtype: (int, int, float)
    """
    # Q-ensemble source: "A faster Horse on a safer trail".
    if seed is not None:
        np.random.seed(seed)
    (i, beta_i) = args_1
    (j, beta_j) = args_2

    q_ensemble = 1/(beta_i + beta_j)
    w_link = np.random.exponential(q_ensemble)
    return (i, j, w_link)


def is_a_link_crema_ecm_prob(args_1, args_2, p_ensemble, seed=None):
    """The function randomly returns an undirected link with related weight
        between two given nodes after the CReMa undirected
        for probabilistic o topology.

    :param args_1: Tuple containing node 1, and strength parameter solution.
    :type args_1: (int, float)
    :param args_2: Tuple containing node 2, and strength parameter solution.
    :type args_2: (int, float)
    :param p_ensemble: Probability the link exists
    :type p_ensemble: float
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: If the links exists returns the triplet of nodes and weight,
        otherwise returns None
    :rtype: (int, int, float)
    """
    # Q-ensemble source: "A faster Horse on a safer trail".
    if seed is not None:
        np.random.seed(seed)
    (i, beta_i) = args_1
    (j, beta_j) = args_2

    p = np.random.random_sample()
    if p < p_ensemble:
        q_ensemble = 1/(beta_i + beta_j)
        w_link = np.random.exponential(q_ensemble)
        return (i, j, w_link)


def is_a_link_crema_sparse_ecm_prob(args_1, args_2, seed=None):
    """The function randomly returns an undirected link with related weight
        between two given nodes after the CReMa undirected
        for probabilistic o topology.
        Function for the sparse version of CReMa.

    :param args_1: Tuple containing node 1, strength parameter solution and
        degree parameter solution from the probabilistic model.
    :type args_1: (int, float)
    :param args_2: Tuple containing node 2, strength parameter solution and
        degree parameter solution from the probabilistic model
    :type args_2: (int, float)
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: If the links exists returns the triplet of nodes and weight,
        otherwise returns None
    :rtype: (int, int, float)
    """
    # Q-ensemble source: "A faster Horse on a safer trail".
    if seed is not None:
        np.random.seed(seed)
    (i, beta_i, x_i) = args_1
    (j, beta_j, x_j) = args_2

    p = np.random.random_sample()
    p_ensemble = x_i*x_j/(1 + x_j*x_i)
    if p < p_ensemble:
        q_ensemble = 1/(beta_i + beta_j)
        w_link = np.random.exponential(q_ensemble)
        return (i, j, w_link)


def is_a_link_crema_decm_det(args_1, args_2, seed=None):
    """The function randomly returns a directed link with related weight
        between two given nodes after the CReMa undirected for a deterministic
         topology.

    :param args_1: Tuple containing out node and out-strength parameter
        solution.
    :type args_1: (int, float)
    :param args_2: Tuple containing in node and in-strength parameter solution.
    :type args_2: (int, float)
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: If the links exists returns the triplet of nodes and weight,
        otherwise returns None
    :rtype: (int, int, float)
    """
    # Q-ensemble source: "A faster Horse on a safer trail".
    if seed is not None:
        np.random.seed(seed)
    (i, b_out_i) = args_1
    (j, b_in_j) = args_2

    q_ensemble = 1/(b_out_i + b_in_j)
    w_link = np.random.exponential(q_ensemble)
    return (i, j, w_link)


def is_a_link_crema_decm_prob(args_1, args_2, p_ensemble, seed=None):
    """The function randomly returns a directed link with related weight
        between two given nodes after the CReMa directed
        for a probabilistic topology.

    :param args_1: Tuple containing out node and out-strength parameter
        solution.
    :type args_1: (int, float)
    :param args_2: Tuple containing in node and in-strength parameter solution.
    :type args_2: (int, float)
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: If the links exists returns the triplet of nodes and weight,
        otherwise returns None
    :rtype: (int, int, float)
    """
    # Q-ensemble source: "A faster Horse on a safer trail".
    if seed is not None:
        np.random.seed(seed)
    (i, b_out_i) = args_1
    (j, b_in_j) = args_2

    p = np.random.random_sample()
    if p < p_ensemble:
        q_ensemble = 1/(b_out_i + b_in_j)
        w_link = np.random.exponential(q_ensemble)
        return (i, j, w_link)


def is_a_link_crema_sparse_decm_prob(args_1, args_2, seed=None):
    """The function randomly returns a directed link with related weight
        between two given nodes after the CReMa directed
        for a probabilistic topology.
        Function for the sparse version of CReMa.

    :param args_1: Tuple containing out node, out-strength parameter solution
        and out-degree parameter solution from the probabilistic model
    :type args_1: (int, float, float)
    :param args_2: Tuple containing in node and in-strength parameter solution
        and in-degree parameter solution from the probabilistic model
    :type args_2: (int, float, float)
    :param seed: Random seed, defaults to None
    :type seed: int, optional
    :return: If the links exists returns the triplet of nodes and weight,
        otherwise returns None
    :rtype: (int, int, float)
    """
    # Q-ensemble source: "A faster Horse on a safer trail".
    if seed is not None:
        np.random.seed(seed)
    (i, b_out_i, x_i) = args_1
    (j, b_in_j, x_j) = args_2

    p = np.random.random_sample()
    p_ensemble = x_i*x_j/(1 + x_j*x_i)
    if p < p_ensemble:
        q_ensemble = 1/(b_out_i + b_in_j)
        w_link = np.random.exponential(q_ensemble)
        return (i, j, w_link)



from .graph_classes import *
from .matrix_generator import *
from .ensemble_generator import *
from .models_functions import *

__version__ = "1.0.5"
__author__ = """Nicolo' Vallarano (nicolo.vallarano@imtlucca.it)
Emiliano Marchese (emiliano.marchese@imtulucca.it)"""



import numpy as np
import scipy.sparse
import scipy
from numba import jit
from scipy.sparse import csr_matrix


def build_graph_from_edgelist(edgelist,
                              is_directed,
                              is_sparse=False,
                              is_weighted=False):
    """Generates adjacency matrix given edgelist.

    :param edgelist: Edgelist.
    :type edgelist: numpy.ndarray
    :param is_directed: True if edge direction is informative.
    :type is_directed: bool
    :param is_sparse: If true the output is a sparse matrix.
    :type is_sparse: bool
    :param is_weighted: True if the edgelist is weighted
    :type is_weighted: bool
    :return: Adjacency matrix.
    :rtype: numpy.ndarray or scipy.sparse.csr_matrix
    """
    if not isinstance(edgelist, np.ndarray):
        raise TypeError("edgelist must be an numpy.ndarray.")
    if is_sparse:
        if is_weighted:
            adjacency = build_graph_sparse_weighted(edgelist, is_directed)
        else:
            adjacency = build_graph_sparse(edgelist, is_directed)
    else:
        if is_weighted:
            adjacency = build_graph_fast_weighted(edgelist, is_directed)
        else:
            adjacency = build_graph_fast(edgelist, is_directed)
    return adjacency


@jit(nopython=True)
def build_graph_fast(edgelist, is_directed):
    """Generates adjacency matrix given edgelist, numpy array format
    is used.

    :param edgelist: Edgelist.
    :type edgelist: numpy.ndarray
    :param is_directed: True if edge direction is informative.
    :type is_directed: bool
    :return: Adjacency matrix.
    :rtype: numpy.ndarray
    """
    if is_directed:
        n_nodes = len(set(edgelist[:, 0]) | set(edgelist[:, 1]))
        adj = np.zeros((n_nodes, n_nodes))
        for edges in edgelist:
            i = int(edges[0])
            j = int(edges[1])
            adj[i, j] = 1
    else:
        n_nodes = len(set(edgelist[:, 0]) | set(edgelist[:, 1]))
        adj = np.zeros((n_nodes, n_nodes))
        for edges in edgelist:
            i = int(edges[0])
            j = int(edges[1])
            adj[i, j] = 1
            adj[j, i] = 1
    return adj


def build_graph_sparse(edgelist, is_directed):
    """Generates adjacency matrix given edgelist, scipy sparse format
    is used.

    :param edgelist: Edgelist.
    :type edgelist: numpy.ndarray
    :param is_sparse: True if edge direction is informative.
    :type is_sparse: bool
    :return: Adjacency matrix.
    :rtype: scipy.sparse.csr_matrix
    """
    n_nodes = set(edgelist[:, 0]) | set(edgelist[:, 1])
    if is_directed:
        row = edgelist[:, 0].astype(int)
        columns = edgelist[:, 1].astype(int)
        data = np.ones(n_nodes, dtype=int)
        adj = csr_matrix((data, (row, columns)), shape=(n_nodes, n_nodes))
    else:
        row = np.concatenate([edgelist[:, 0], edgelist[:, 1]]).astype(int)
        columns = np.concatenate([edgelist[:, 1], edgelist[:, 0]]).astype(int)
        data = np.ones(2*n_nodes, dtype=int)
        adj = csr_matrix((data, (row, columns)), shape=(n_nodes, n_nodes))
    return adj


@jit(nopython=True)
def build_graph_fast_weighted(edgelist, is_directed):
    """Generates weighted adjacency matrix given edgelist,
    numpy array format is used.

    :param edgelist: Edgelist.
    :type edgelist: numpy.ndarray
    :param is_directed: True if edge direction is informative.
    :type is_directed: bool
    :return: Adjacency matrix.
    :rtype: numpy.ndarray
    """
    if is_directed:
        n_nodes = len(set(edgelist[:, 0]) | set(edgelist[:, 1]))
        adj = np.zeros((n_nodes, n_nodes))
        for edges in edgelist:
            i = int(edges[0])
            j = int(edges[1])
            w = edges[2]
            adj[i, j] = w
    else:
        n_nodes = len(set(edgelist[:, 0]) | set(edgelist[:, 1]))
        adj = np.zeros((n_nodes, n_nodes))
        for edges in edgelist:
            i = int(edges[0])
            j = int(edges[1])
            w = edges[2]
            adj[i, j] = w
            adj[j, i] = w
    return adj


def build_graph_sparse_weighted(edgelist, is_directed):
    """Generates weighted adjacency matrix given edgelist,
    scipy sparse format is used.

    :param edgelist: Edgelist.
    :type edgelist: numpy.ndarray
    :param is_sparse: True if edge direction is informative.
    :type is_sparse: bool
    :return: Adjacency matrix.
    :rtype: scipy.sparse.csr_matrix
    """
    n_nodes = set(edgelist[:, 0]) | set(edgelist[:, 1])
    if is_directed:
        row = edgelist[:, 0].astype(int)
        columns = edgelist[:, 1].astype(int)
        data = edgelist[:, 2]
        adj = csr_matrix((data, (row, columns)), shape=(n_nodes, n_nodes))
    else:
        row = np.concatenate([edgelist[:, 0], edgelist[:, 1]]).astype(int)
        columns = np.concatenate([edgelist[:, 1], edgelist[:, 0]]).astype(int)
        data = np.concatenate([edgelist[:, 2], edgelist[:, 2]])
        adj = csr_matrix((data, (row, columns)), shape=(n_nodes, n_nodes))
    return adj



def out_degree(a):
    """Returns matrix *a* out degrees sequence.

    :param a: Adjacency matrix
    :type a: numpy.ndarray, scipy.sparse.csr.csr_matrix,
        scipy.sparse.coo.coo_matrix
    :return: Out degree sequence.
    :rtype: numpy.ndarray
    """
    # if the matrix is a numpy array
    if type(a) == np.ndarray:
        return np.sum(a > 0, 1)
    # if the matrix is a scipy sparse matrix
    elif type(a) in [scipy.sparse.csr.csr_matrix, scipy.sparse.coo.coo_matrix]:
        return np.sum(a > 0, 1).A1  # noqa


def in_degree(a):
    """Returns matrix *a* in degrees sequence.

    :param a: Adjacency matrix.
    :type a: numpy.ndarray, scipy.sparse.csr.csr_matrix,
        scipy.sparse.coo.coo_matrix
    :return: In degree sequence.
    :rtype: numpy.ndarray
    """
    # if the matrix is a numpy array
    if type(a) == np.ndarray:
        return np.sum(a > 0, 0)
    # if the matrix is a scipy sparse matrix
    elif type(a) in [scipy.sparse.csr.csr_matrix, scipy.sparse.coo.coo_matrix]:
        return np.sum(a > 0, 0).A1


def out_strength(a):
    """Returns matrix *a* out strengths sequence.

    :param a: Adjacency matrix.
    :type a: numpy.ndarray, scipy.sparse.csr.csr_matrix,
        scipy.sparse.coo.coo_matrix
    :return: Out strengths sequence.
    :rtype: numpy.ndarray
    """
    # if the matrix is a numpy array
    if type(a) == np.ndarray:
        return np.sum(a, 1)
    # if the matrix is a scipy sparse matrix
    elif type(a) in [scipy.sparse.csr.csr_matrix, scipy.sparse.coo.coo_matrix]:
        return np.sum(a, 1).A1


def in_strength(a):
    """Returns matrix *a* in strengths sequence.

    :param a: Adjacency matrix.
    :type a: numpy.ndarray, scipy.sparse.csr.csr_matrix,
        scipy.sparse.coo.coo_matrix
    :return: In strengths sequence.
    :rtype: numpy.ndarray
    """
    # if the matrix is a numpy array
    if type(a) == np.ndarray:
        return np.sum(a, 0)
    # if the matrix is a scipy sparse matrix
    elif type(a) in [scipy.sparse.csr.csr_matrix, scipy.sparse.coo.coo_matrix]:
        return np.sum(a, 0).A1


def degree(a):
    """Returns matrix *a* degrees sequence.

    :param a: Adjacency matrix.
    :type a: numpy.ndarray, scipy.sparse
    :return: Degree sequence.
    :rtype: numpy.ndarray
    """
    # if the matrix is a numpy array
    if type(a) == np.ndarray:
        return np.sum(a > 0, 1)
    # if the matrix is a scipy sparse matrix
    elif type(a) in [scipy.sparse.csr.csr_matrix, scipy.sparse.coo.coo_matrix]:
        return np.sum(a > 0, 1).A1


def strength(a):
    """Returns matrix *a* strengths sequence.

    :param a: Adjacency matrix.
    :type a: numpy.ndarray, scipy.sparse
    :return: Strengths sequence.
    :rtype: numpy.ndarray
    """
    # if the matrix is a numpy array
    if type(a) == np.ndarray:
        return np.sum(a, 1)
    # if the matrix is a scipy sparse matrix
    elif type(a) in [scipy.sparse.csr.csr_matrix, scipy.sparse.coo.coo_matrix]:
        return np.sum(a, 1).A1


def edgelist_from_edgelist_undirected(edgelist):
    """Creates a new edgelist with the indexes of the nodes instead of the names. Returns also a dictionary that keep track of the nodes and, depending on the type of graph, degree and strengths sequences.

    :param edgelist: edgelist.
    :type edgelist: numpy.ndarray or list
    :return: edgelist, degrees sequence, strengths sequence and new labels to old labels dictionary.
    :rtype: (numpy.ndarray, numpy.ndarray, numpy.ndarray, dict)
    """
    edgelist = [tuple(item) for item in edgelist]
    if len(edgelist[0]) == 2:
        nodetype = type(edgelist[0][0])
        edgelist = np.array(
            edgelist,
            dtype=np.dtype([("source", nodetype), ("target", nodetype)]),
        )
    else:
        nodetype = type(edgelist[0][0])
        weigthtype = type(edgelist[0][2])
        # Vorrei mettere una condizione sul weighttype che deve essere numerico
        edgelist = np.array(
            edgelist,
            dtype=np.dtype(
                [
                    ("source", nodetype),
                    ("target", nodetype),
                    ("weigth", weigthtype),
                ]
            ),
        )
    # If there is a loop we count it twice in the degree of the node.
    unique_nodes, degree_seq = np.unique(
        np.concatenate((edgelist["source"], edgelist["target"])),
        return_counts=True,
    )
    nodes_dict = dict(enumerate(unique_nodes))
    inv_nodes_dict = {v: k for k, v in nodes_dict.items()}
    if len(edgelist[0]) == 2:
        edgelist_new = [
            (inv_nodes_dict[edge[0]], inv_nodes_dict[edge[1]])
            for edge in edgelist
        ]
        edgelist_new = np.array(
            edgelist_new, dtype=np.dtype([("source", int), ("target", int)])
        )
    else:
        edgelist_new = [
            (inv_nodes_dict[edge[0]], inv_nodes_dict[edge[1]], edge[2])
            for edge in edgelist
        ]
        edgelist_new = np.array(
            edgelist_new,
            dtype=np.dtype(
                [("source", int), ("target", int), ("weigth", weigthtype)]
            ),
        )
    if len(edgelist[0]) == 3:
        aux_edgelist = np.concatenate(
            (edgelist_new["source"], edgelist_new["target"])
        )
        aux_weights = np.concatenate(
            (edgelist_new["weigth"], edgelist_new["weigth"])
        )
        strength_seq = np.array(
            [aux_weights[aux_edgelist == i].sum() for i in unique_nodes]
        )
        return edgelist_new, degree_seq, strength_seq, nodes_dict
    return edgelist_new, degree_seq, nodes_dict


def edgelist_from_edgelist_directed(edgelist):
    """Creates a new edgelists replacing nodes labels with indexes.
    Returns also two dictionaries that keep track of the
    nodes index-label relation.
    Works also on weighted graphs.

    :param edgelist: List of edges.
    :type edgelist: list
    :return: Re-indexed list of edges, out-degrees, in-degrees,
        index to label dictionary
    :rtype: (dict, numpy.ndarray, numpy.ndarray, dict)
    """
    # TODO: inserire esempio edgelist pesata edgelist binaria
    # nel docstring
    # edgelist = list(zip(*edgelist))
    if len(edgelist[0]) == 2:
        nodetype = type(edgelist[0][0])
        edgelist = np.array(
            edgelist,
            dtype=np.dtype(
                [
                    ("source", nodetype),
                    ("target", nodetype)
                ]
            ),
        )
    else:
        nodetype = type(edgelist[0][0])
        weigthtype = type(edgelist[0][2])
        # Vorrei mettere una condizione sul weighttype che deve essere numerico
        edgelist = np.array(
            edgelist,
            dtype=np.dtype(
                [
                    ("source", nodetype),
                    ("target", nodetype),
                    ("weigth", weigthtype),
                ]
            ),
        )
    # If there is a loop we count it twice in the degree of the node.
    unique_nodes = np.unique(
        np.concatenate((edgelist["source"], edgelist["target"])),
        return_counts=False,
    )
    out_degree = np.zeros_like(unique_nodes)
    in_degree = np.zeros_like(unique_nodes)
    nodes_dict = dict(enumerate(unique_nodes))
    inv_nodes_dict = {v: k for k, v in nodes_dict.items()}
    if len(edgelist[0]) == 2:
        edgelist_new = [
            (inv_nodes_dict[edge[0]], inv_nodes_dict[edge[1]])
            for edge in edgelist
        ]
        edgelist_new = np.array(
            edgelist_new, dtype=np.dtype([("source", int), ("target", int)])
        )
    else:
        edgelist_new = [
            (inv_nodes_dict[edge[0]], inv_nodes_dict[edge[1]], edge[2])
            for edge in edgelist
        ]
        edgelist_new = np.array(
            edgelist_new,
            dtype=np.dtype(
                [("source", int), ("target", int), ("weigth", weigthtype)]
            ),
        )
    out_indices, out_counts = np.unique(
        edgelist_new["source"], return_counts=True
    )
    in_indices, in_counts = np.unique(
        edgelist_new["target"], return_counts=True
    )
    out_degree[out_indices] = out_counts
    in_degree[in_indices] = in_counts
    if len(edgelist[0]) == 3:
        out_strength = np.zeros_like(unique_nodes, dtype=weigthtype)
        in_strength = np.zeros_like(unique_nodes, dtype=weigthtype)
        out_counts_strength = np.array(
            [
                edgelist_new[edgelist_new["source"] == i]["weigth"].sum()
                for i in out_indices
            ]
        )
        in_counts_strength = np.array(
            [
                edgelist_new[edgelist_new["target"] == i]["weigth"].sum()
                for i in in_indices
            ]
        )
        out_strength[out_indices] = out_counts_strength
        in_strength[in_indices] = in_counts_strength
        return (
            edgelist_new,
            out_degree,
            in_degree,
            out_strength,
            in_strength,
            nodes_dict,
        )
    return edgelist_new, out_degree, in_degree, nodes_dict

# Bipartite networks functions


@jit(nopython=True)
def bicm_from_fitnesses(x, y):
    """
    Rebuilds the average probability matrix of the bicm2 from the fitnesses

    :param x: the fitness vector of the rows layer
    :type x: numpy.ndarray
    :param y: the fitness vector of the columns layer
    :type y: numpy.ndarray
    """
    avg_mat = np.zeros((len(x), len(y)))
    for i in range(len(x)):
        for j in range(len(y)):
            xy = x[i] * y[j]
            avg_mat[i, j] = xy / (1 + xy)
    return avg_mat


def sample_bicm(avg_mat):
    """
    Build a biadjacency matrix sampling from the probability matrix of a BiCM.
    """
    if not isinstance(avg_mat, np.ndarray):
        avg_mat = np.array(avg_mat)
    dim1, dim2 = avg_mat.shape
    return np.array(avg_mat > np.reshape(np.random.sample(dim1 * dim2), (dim1, dim2)), dtype=int)


def sample_bicm_edgelist(x, y):
    """
    Build an edgelist sampling from the fitnesses of a BiCM.
    """
    edgelist = []
    for i in range(len(x)):
        for j in range(len(y)):
            xy = x[i] * y[j]
            if np.random.uniform() < xy / (1 + xy):
                edgelist.append((i, j))
    return edgelist


def sample_bicm_edgelist_names(dict_x, dict_y):
    """
    Build an edgelist from the BiCM keeping the names of the nodes as contained in the BiCM fitnesses' dictionaries.
    """
    edgelist = []
    for xx in dict_x:
        for yy in dict_y:
            xy = dict_x[xx] * dict_y[yy]
            if np.random.uniform() < xy / (1 + xy):
                edgelist.append((xx, yy))
    return edgelist


@jit(nopython=True)
def edgelist_from_biadjacency_fast(biadjacency):
    """
    Build the edgelist of a bipartite network from its biadjacency matrix.
    """
    edgelist = []
    for i in range(biadjacency.shape[0]):
        for j in range(biadjacency.shape[1]):
            if biadjacency[i, j] == 1:
                edgelist.append((i, j))
    return edgelist


def edgelist_from_biadjacency(biadjacency):
    """
    Build the edgelist of a bipartite network from its biadjacency matrix.
    Accounts for sparse matrices and returns a structured array.
    """
    if scipy.sparse.isspmatrix(biadjacency):
        coords = biadjacency.nonzero()
        if np.sum(biadjacency.data != 1) > 0:
            raise ValueError('Only binary matrices')
        return np.array(list(zip(coords[0], coords[1])), dtype=np.dtype([('rows', int), ('columns', int)])),\
               np.array(biadjacency.sum(1)).flatten(), np.array(biadjacency.sum(0)).flatten()
    else:
        if np.sum(biadjacency[biadjacency != 0] != 1) > 0:
            raise ValueError('Only binary matrices')
        return np.array(edgelist_from_biadjacency_fast(biadjacency),
                        dtype=np.dtype([('rows', int), ('columns', int)])),\
               np.sum(biadjacency, axis=1), np.sum(biadjacency, axis=0)


def biadjacency_from_edgelist(edgelist, fmt='array'):
    """
    Build the biadjacency matrix of a bipartite network from its edgelist.
    Returns a matrix of the type specified by ``fmt``, by default a numpy array.
    """
    edgelist, rows_deg, cols_deg, rows_dict, cols_dict = edgelist_from_edgelist_bipartite(edgelist)
    if fmt == 'array':
        biadjacency = np.zeros((len(rows_deg), len(cols_deg)), dtype=int)
        for edge in edgelist:
            biadjacency[edge[0], edge[1]] = 1
    elif fmt == 'sparse':
        biadjacency = scipy.sparse.coo_matrix((np.ones(len(edgelist)), (edgelist['rows'], edgelist['columns'])))
    elif not isinstance(format, str):
        raise TypeError('format must be a string (either "array" or "sparse")')
    else:
        raise ValueError('format must be either "array" or "sparse"')
    return biadjacency, rows_deg, cols_deg, rows_dict, cols_dict


def edgelist_from_edgelist_bipartite(edgelist):
    """
    Creates a new edgelist with the indexes of the nodes instead of the names.
    Method for bipartite networks.
    Returns also two dictionaries that keep track of the nodes.
    """
    edgelist = np.array(list(set([tuple(edge) for edge in edgelist])))
    out = np.zeros(np.shape(edgelist)[0], dtype=np.dtype([('source', object), ('target', object)]))
    out['source'] = edgelist[:, 0]
    out['target'] = edgelist[:, 1]
    edgelist = out
    unique_rows, rows_degs = np.unique(edgelist['source'], return_counts=True)
    unique_cols, cols_degs = np.unique(edgelist['target'], return_counts=True)
    rows_dict = dict(enumerate(unique_rows))
    cols_dict = dict(enumerate(unique_cols))
    inv_rows_dict = {v: k for k, v in rows_dict.items()}
    inv_cols_dict = {v: k for k, v in cols_dict.items()}
    edgelist_new = [(inv_rows_dict[edge[0]], inv_cols_dict[edge[1]]) for edge in edgelist]
    edgelist_new = np.array(edgelist_new, dtype=np.dtype([('rows', int), ('columns', int)]))
    return edgelist_new, rows_degs, cols_degs, rows_dict, cols_dict


def adjacency_list_from_edgelist_bipartite(edgelist, convert_type=True):
    """
    Creates the adjacency list from the edgelist.
    Method for bipartite networks.
    Returns two dictionaries, each containing an adjacency list with the rows as keys and the columns as keys, respectively.
    If convert_type is True (default), then the nodes are enumerated and the adjacency list is returned as integers.
    Returns also two dictionaries that keep track of the nodes and the two degree sequences.
    """
    if convert_type:
        edgelist, rows_degs, cols_degs, rows_dict, cols_dict = edgelist_from_edgelist_bipartite(edgelist)
    adj_list = {}
    inv_adj_list = {}
    for edge in edgelist:
        adj_list.setdefault(edge[0], set()).add(edge[1])
        inv_adj_list.setdefault(edge[1], set()).add(edge[0])
    if not convert_type:
        rows_degs = np.array([len(adj_list[k]) for k in adj_list])
        rows_dict = {k: k for k in adj_list}
        cols_degs = np.array([len(inv_adj_list[k]) for k in inv_adj_list])
        cols_dict = {k: k for k in inv_adj_list}
    return adj_list, inv_adj_list, rows_degs, cols_degs, rows_dict, cols_dict


def adjacency_list_from_adjacency_list_bipartite(old_adj_list):
    """
    Creates the adjacency list from another adjacency list, convering the data type.
    Method for bipartite networks.
    Returns two dictionaries, each containing an adjacency list with the rows as keys and the columns as keys, respectively.
    Original keys are treated as rows, values as columns.
    The nodes are enumerated and the adjacency list is returned as integers.
    Returns also two dictionaries that keep track of the nodes and the two degree sequences.
    """
    rows_dict = dict(enumerate(np.unique(list(old_adj_list.keys()))))
    cols_dict = dict(enumerate(np.unique([el for mylist in old_adj_list.values() for el in mylist])))
    inv_rows_dict = {v: k for k, v in rows_dict.items()}
    inv_cols_dict = {v: k for k, v in cols_dict.items()}
    adj_list = {}
    inv_adj_list = {}
    for k in old_adj_list:
        adj_list.setdefault(inv_rows_dict[k], set()).update({inv_cols_dict[val] for val in old_adj_list[k]})
        for val in old_adj_list[k]:
            inv_adj_list.setdefault(inv_cols_dict[val], set()).add(inv_rows_dict[k])
    rows_degs = np.array([len(adj_list[k]) for k in adj_list])
    cols_degs = np.array([len(inv_adj_list[k]) for k in inv_adj_list])
    return adj_list, inv_adj_list, rows_degs, cols_degs, rows_dict, cols_dict


def adjacency_list_from_biadjacency(biadjacency):
    """
    Creates the adjacency list from a biadjacency matrix, given in sparse format or as a list or numpy array.
    Returns two dictionaries, each containing an adjacency list with the rows as keys and the columns as keys, respectively.
    Returns also the two degree sequences.
    """
    if scipy.sparse.isspmatrix(biadjacency):
        if np.sum(biadjacency.data != 1) > 0:
            raise ValueError('Only binary matrices')
        coords = biadjacency.nonzero()
    else:
        biadjacency = np.array(biadjacency)
        if np.sum(biadjacency[biadjacency != 0] != 1) > 0:
            raise ValueError('Only binary matrices')
        coords = np.where(biadjacency != 0)
    adj_list = {}
    inv_adj_list = {}
    for edge_i in range(len(coords[0])):
        adj_list.setdefault(coords[0][edge_i], set()).add(coords[1][edge_i])
        inv_adj_list.setdefault(coords[1][edge_i], set()).add(coords[0][edge_i])
    rows_degs = np.array([len(adj_list[k]) for k in adj_list])
    cols_degs = np.array([len(inv_adj_list[k]) for k in inv_adj_list])
    return adj_list, inv_adj_list, rows_degs, cols_degs
